module Main exposing (..)

import Html exposing (..)
import Html.Attributes exposing (..)
import Browser
import Html.Events exposing (..)
import List
import String
import Task exposing (..)
import Time exposing (Posix)

import Date exposing (..)


{-
1. NOTE:
    The Date.sub implementation provided is broken. It will not terminate given specific parameters, such as (29, 2, 2022) (21, 12, 2022).
    We have replaced it with our own implementation seen below, which is also more efficient, running in O(1) compared to O(n):

    sub : Date -> Date -> Int
    sub date1 date2 = convertDateToDayNumber date1 - convertDateToDayNumber date2

    -- https://web.archive.org/web/20170507133619/https://alcor.concordia.ca/~gpkatch/gdate-algorithm.html
    convertDateToDayNumber : Date -> Int
    convertDateToDayNumber (a, b, c) =
        let m = (modBy 12 (b + 9))
            y = c - m // 10
        in 365 * y + y // 4 - y // 100 + y // 400 + (m * 306 + 5) // 10 + ( a - 1 )

    convertDayNumberToDate : Int -> Date
    convertDayNumberToDate g =
        let y = (10000 * g + 14780) // 3652425
            ddd = g - (365 * y + y // 4 - y // 100 + y // 400)
        in
            if (ddd < 0) then
                let y2 = y - 1 -- y = y2
                    ddd2 = g - (365 * y2 + y2 // 4 - y2 // 100 + y2 // 400)  -- ddd = ddd2
                    mi = (100 * ddd2 + 52) // 3060
                    mm = (modBy (12) (mi + 2)) + 1
                    y3 = y2 + (mi + 2) // 12  -- y = y3
                    dd = ddd2 - (mi * 306 + 5) // 10 + 1
                in (dd, mm, y3)
            else
                let mi = (100 * ddd + 52) // 3060
                    mm = (modBy (12) (mi + 2)) + 1
                    y2 = y + (mi + 2) // 12  -- y = y2
                    dd = ddd - (mi * 306 + 5) // 10 + 1
                in (dd, mm, y2)

2. NOTE:
   We had to move the elm/time dependency from "indirect" to "direct" in order to be able to use the Time.Posix, Time.Zone and Time.Month types.
   With that, our dependencies are the following:
   "dependencies": {
       "direct": {
           "elm/browser": "1.0.2",
           "elm/core": "1.0.5",
           "elm/html": "1.0.0",
           "elm/time": "1.0.0"
       },
       "indirect": {
           "elm/json": "1.1.3",
           "elm/url": "1.0.0",
           "elm/virtual-dom": "1.0.3"
       }
   }
-}


type alias Model =
    {
    arbeit : Arbeit
    , beginndatum : Date
    , besprechungsWochentag : Weekday
    , anzahlWochenZwischenBesprechung : Int
    , status : String
    , zone : Time.Zone
    , time : Time.Posix
    }


type Msg
    = SetStartDate String
    | SetInterval String
    | SetDegree String
    | SetBesprechundsWochentag String
    | Tick Posix
    | AdjustTimeZone Time.Zone

type Arbeit
    = Bachelor
    | Master

init : () -> (Model, Cmd Msg)
init _ = (
    {
        arbeit = Bachelor
        , beginndatum = (21, 12, 2022)
        , besprechungsWochentag = Mon
        , anzahlWochenZwischenBesprechung = 1
        , status = ""
        , zone = Time.utc
        , time = Time.millisToPosix 0
    } , Task.perform AdjustTimeZone Time.here
    )


addMonths : Int -> Date -> Date
addMonths count date =
    if count > 0
    then addMonths (count - 1) (addDays (getMonthLength date) date)
    else date

addDays : Int -> Date -> Date
addDays count date =
    if count > 0
    then addDays (count - 1) (Date.next date)
    else date

getWeekday : Date -> Weekday
getWeekday date =
        addWeekdays (Date.sub date (21, 12, 2022)) Wed

addWeekdays : Int -> Weekday -> Weekday
addWeekdays count day =
    if count < 0 then
        case count of
            0 -> day
            _ -> addWeekdays (count + 1) (prevWeekDay day)
    else if count > 0 then
        case count of
            0 -> day
            _ -> addWeekdays (count - 1) (nextWeekDay day)
    else day

nextWeekDay : Weekday -> Weekday
nextWeekDay day =
    case day of
        Mon -> Tue
        Tue -> Wed
        Wed -> Thu
        Thu -> Fri
        Fri -> Sat
        Sat -> Sun
        Sun -> Mon

prevWeekDay : Weekday -> Weekday
prevWeekDay day =
    case day of
        Mon -> Sun
        Tue -> Mon
        Wed -> Tue
        Thu -> Wed
        Fri -> Thu
        Sat -> Fri
        Sun -> Sat

getMonthLength : Date -> Int
getMonthLength date =
    case Date.month date of
        1 -> 31
        2 -> if Date.leapyear (Date.year date) then 28 else 29
        3 -> 31
        4 -> 30
        5 -> 31
        6 -> 30
        7 -> 31
        8 -> 31
        9 -> 30
        10 -> 31
        11 -> 30
        12 -> 31
        _ -> 30

{- will increment the month count by the given amount. Supports a max of 12 months to be added. Will limit the day to
   the amount of days the new month has. -}
addMonthsLiterally : Int -> Date -> Date
addMonthsLiterally count date =
    let newMonth = (Date.month date) + count
        newDate = (Date.day date, if newMonth > 12 then (modBy 12 newMonth) else newMonth, if newMonth > 12 then (Date.year date) + 1 else Date.year date)
        newDateMonthLength = getMonthLength newDate
    in if Date.day newDate > newDateMonthLength then (Date.day date, newDateMonthLength, Date.year date) else newDate

skipWeekend : Date -> Date
skipWeekend date =
    case getWeekday date of
        Sat -> addDays 2 date
        Sun -> addDays 1 date
        _ -> date

{- will advance the date so long until all days in [Mon, Tue, Wed, Thu, Fri] are covered.
   This does not include the first day -}
advanceByEntireWorkWeek : Date -> Date
advanceByEntireWorkWeek date =
    date
    |> skipDaysWhilstSkippingWeekend 5
    |> skipWeekend

skipDaysWhilstSkippingWeekend : Int -> Date -> Date
skipDaysWhilstSkippingWeekend count date =
    if count <= 0 then date else
        skipDaysWhilstSkippingWeekend (count - 1) (date
                                                   |> skipWeekend
                                                   |> Date.next)

formatDateOutput : Date -> String
formatDateOutput date =
    String.fromInt (Date.day date) ++ "." ++ String.fromInt (Date.month date) ++ "." ++ String.fromInt (Date.year date)

calculateAbgabeEndDate : Arbeit -> Date -> Date
calculateAbgabeEndDate arbeit date =
    case arbeit of
        Bachelor -> Date.prev (addMonthsLiterally 3 date)
        Master -> Date.prev (addMonthsLiterally 6 date)


validateModel : Model -> Result String String
validateModel model =
    case model.besprechungsWochentag of
        Sat -> Err "Datum darf kein Samstag sein"
        Sun -> Err "Datum darf kein Sonntag sein"
        _ -> case model.beginndatum of
            (day, month, year) -> if day > getMonthLength (day, month, year) || day <= 0 then
                                      Err "Tag ist außerhalb des Monats"
                                  else if month < 1 || month > 12 then
                                           Err "Monat muss zwischen 1 und 12 liegen"
                                       else if year < 1000 || year > 9999 then
                                               Err "Jahr muss größer als 1000 und kleiner als 9999 sein"
                                           else Ok ""

findNextDateOfWeekday : Date -> Weekday -> Date
findNextDateOfWeekday date weekday =
    if getWeekday date == weekday then date else findNextDateOfWeekday (Date.next date) weekday

berechneDaten : Model -> List (String, String)
berechneDaten model =
    let beginndatum = model.beginndatum
        enddatum = calculateAbgabeEndDate model.arbeit beginndatum
        besprechungsWochentag = model.besprechungsWochentag
        anzahlWochenZwischenBesprechung = model.anzahlWochenZwischenBesprechung
    in List.concat [
                      ((formatDateOutput beginndatum), "Beginn") :: (berechneZwischentermine (findNextDateOfWeekday (Date.next beginndatum) besprechungsWochentag) enddatum anzahlWochenZwischenBesprechung besprechungsWochentag)
                    , [((formatDateOutput enddatum), "Ende")]
                   ]

berechneZwischentermine : Date -> Date -> Int -> Weekday -> List (String, String)
berechneZwischentermine datum enddatum wochenAbstand wochentag =
    let nextBesprechung = addDays (wochenAbstand * 7) datum
        isOver = Date.gt datum enddatum || Date.eq datum enddatum
    in if isOver then [] else
        ((formatDateOutput datum), "Besprechung") :: (berechneZwischentermine nextBesprechung enddatum wochenAbstand wochentag)

monthToIndex : Time.Month -> Int
monthToIndex month =
    case month of
        Time.Jan -> 1
        Time.Feb -> 2
        Time.Mar -> 3
        Time.Apr -> 4
        Time.May -> 5
        Time.Jun -> 6
        Time.Jul -> 7
        Time.Aug -> 8
        Time.Sep -> 9
        Time.Oct -> 10
        Time.Nov -> 11
        Time.Dec -> 12

getCurrentDate : Model -> Date
getCurrentDate model =
    (
      Time.toDay model.zone model.time
    , monthToIndex (Time.toMonth model.zone model.time)
    , Time.toYear model.zone model.time
    )

update : Msg -> Model -> (Model, Cmd Msg)
update msg model =
    let newModel = case msg of
            SetStartDate date ->
                {model | beginndatum = (
                    case (String.split " " date) of
                        ["asap"] -> advanceByEntireWorkWeek (getCurrentDate model)
                        ["ASAP"] -> advanceByEntireWorkWeek (getCurrentDate model)
                        [d, m, y] -> case [String.toInt d, String.toInt m, String.toInt y] of
                                        [Just a, Just b, Just c] -> (a, b, c)
                                        _ -> (21, 12, 2022)
                        _ -> (21, 12, 2022)
                ) }
            SetDegree arbeits ->
                case arbeits of
                    "Bachelor" -> {model | arbeit = Bachelor}
                    "Master" -> {model | arbeit = Master}
                    _ -> model
            SetInterval str ->
                {model | anzahlWochenZwischenBesprechung = String.toInt str |> Maybe.withDefault 0}
            SetBesprechundsWochentag weekday ->
                {model | besprechungsWochentag = case weekday of
                                                    "Montag" -> Mon
                                                    "Dienstag" -> Tue
                                                    "Mittwoch" -> Wed
                                                    "Donnerstag" -> Thu
                                                    "Freitag" -> Fri
                                                    _ -> Mon
                }
            Tick newTime -> { model | time = newTime }
            AdjustTimeZone newZone -> { model | zone = newZone }
    in case validateModel newModel of
            Ok a -> ({ newModel | status = case msg of
                                               Tick i -> newModel.status
                                               AdjustTimeZone i -> newModel.status
                                               _ -> a
                     }, Cmd.none)
            Err a -> ({ model | status = a }, Cmd.none)


view : Model -> Html Msg
view model =
    div []
    [
      div [style "display" "flex", style "justify-content" "center", style "margin-top" "20px", style "gap" "5px"]
      [
            input [onInput SetStartDate, placeholder "21 12 2022"] []
          , select [onInput SetDegree, style "" ""]
                    [
                        option [value "Bachelor"] [text "Bachelor"]
                        , option [value "Master"] [text "Master"]
                    ]
          , select [onInput SetBesprechundsWochentag]
          [
                option [value "Montag"] [text "Montag"]
              , option [value "Dienstag"] [text "Dienstag"]
              , option [value "Mittwoch"] [text "Mittwoch"]
              , option [value "Donnerstag"] [text "Donnerstag"]
              , option [value "Freitag"] [text "Freitag"]
          ]
          , select [onInput SetInterval]
          [
                    option [value "1"] [text "1 Woche"]
                  , option [value "2"] [text "2 Wochen"]
                  , option [value "3"] [text "3 Wochen"]
                  , option [value "4"] [text "4 Wochen"]

          ]
      ]
    , br [] []
    , div[style "text-align" "center"] [
        span[style "color" "crimson"] [text model.status]
      ]

    , div [style "display" "flex", style "justify-content" "center", style "margin-top" "20px", style "gap" "5px"] [
         table [style "border-collapse" "collapse", style "width" "max(20%, 200px)"]
            [
                thead [style "text-align" "left", style "border-bottom" "2px solid black"] [
                    tr [] [
                          th [] [text "Datum"]
                        , th [] [text "Typ"]
                    ]
                ]
                , tbody []
                    (model
                     |> berechneDaten
                     |> List.map (\(datum, typ) -> tr [] [td [] [text datum], td [] [text typ]]))
            ]
      ]
    ]

subscriptions : Model -> Sub Msg
subscriptions model =
  Time.every 1000 Tick

main = Browser.element {
          init = init
        , update = update
        , view = view
        , subscriptions = subscriptions
    }

